;;; -*- Mode: LISP; Package: :cl-user; BASE: 10; Syntax: ANSI-Common-Lisp; -*-
;;;
;;;   Touched: Sat Oct 14 14:03:50 2006 +0530 <madhu@edu.unm.cs>
;;;   Time-stamp: <06/10/16 01:10:08 madhu>
;;;   Bugs-To: <madhu@edu.unm.cs>
;;;   Copyright (C) 2006 S.Madhu.
;;;
;;; S.Madhu hereby grants you the rights to distribute and use this
;;; software as governed by the terms of the Lisp Lesser GNU Public
;;; License (http://opensource.franz.com/preamble.html), known as the
;;; LLGPL.
;;;
;;; This software is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
;;;
;;;
;;; OPTIONSMENU
;;;
;;; SYNOPSIS
;;;
;;; Provide the functionality of Motif option menus to select one of
;;; many values. See tk_optionMenu for background.
;;;
;;; ?The functionality may be similar to that of `combo box'?
;;;
;;; An OPTIONSMENU menu widget is created by the function
;;; TK-OPTION-MENU. See its function documentation for details and
;;; EXAMPLE-1 below for sample use. EXAMPLE-2 shows how an OPTIONSMENU
;;; can be used in conjunction with LTK:MENURADIOBUTTONs. Along with a
;;; few extensions, EXAMPLE-3 demonstrates how one can modify the
;;; displayed menuentries by manipulating LTK:MENURADIOBUTTON slots.
;;;
;;;
;;; NOTES
;;;
;;; In Tk's parlance. a `menu' must be a child of a `menubutton'.
;;;
;;; In LTK's parlance when we add LTK:MENUBUTTONs to an LTK:MENU, we
;;; are, in Tk's parlance, adding `menuentries' to a `menu'.
;;;
;;; To work around these disparate notions we introduce an
;;; intermediate parent, a tk-menubutton LTK:Widget, which becomes the
;;; parent of the OPTIONSMENU.
;;;
;;;
;;; CAVEATS
;;;
;;; Use TK-OPTION-MENU to create an instance. Do not call MAKE-INSTANCE
;;; directly.
;;;
;;; This code depends on internal symbols the LTK package was tested
;;; with ltk-0.90.
;;;
(defpackage "TK-OPTIONSMENU-VERSION-0.2"
  (:use "CL")
  (:export "TK-OPTION-MENU"))
(in-package "TK-OPTIONSMENU-VERSION-0.2")


(defclass optionsmenu (ltk:widget)
  ((var :initarg :variable :accessor optionsmenu-variable :initform nil)))

(defmethod initialize-instance :after ((optionsmenu optionsmenu) &key
				       &allow-other-keys)
  (unless (optionsmenu-variable optionsmenu)
    (setf (optionsmenu-variable optionsmenu)
	  (ltk::name optionsmenu))))


(defmethod ltk:value ((optionsmenu optionsmenu))
  (let ((var (optionsmenu-variable optionsmenu)))
    (when var
      (ltk:format-wish "senddatastring $~a" var)
      (ltk::read-data))))

(defmethod (setf ltk:value) (val (optionsmenu optionsmenu))
  (ltk:format-wish "set ~a \"~a\"" (optionsmenu-variable optionsmenu)
		   (ltk::tkescape val))
  val)


(defun tk-option-menu (&key ((:master parent)) variable underline options
		       (selected-index 0))
  "Create and returns an OPTIONSMENU widget which then provides
the functionality of Motif option menus for selecting one of the
many OPTIONS through a dropdown menu.

OPTIONS are a list of strings. SELECTED-INDEX is an integer index
into the OPTIONS list. The OPTIONS are available for selection as
radiobutton menuentries. The value of the currently selected
option is available via the LTK:VALUE accessor.

Internally the function also creates an LTK:WIDGET representing
the Tk-menubutton which the OPTIONSMENU menu underlies. Note this
Tk menubutton widget is available via the LTK:MASTER accessor and
can be configured separately. Calls to LTK:PACK are forwarded to
this Tk-menubutton.

The value of the currently selected option is displayed as the
label of this Tk-menubutton via a Tk textvariable named by
VARIABLE. If VARIABLE is NIL, its value defaults to the name of
the OPTIONSMENU widget.
"
  (let* ((tk-menubutton (make-instance 'ltk:widget :master parent))
	 (optionsmenu (make-instance 'optionsmenu :master tk-menubutton
				     :variable variable))
	 (group (optionsmenu-variable optionsmenu)))
    (ltk:format-wish "menubutton ~A -textvariable ~A -indicatoron 1 -menu ~A -relief raised -borderwidth 2 -highlightthickness 2 -anchor c -direction flush"
		     (ltk:widget-path tk-menubutton) group
		     (ltk:widget-path optionsmenu))
    (ltk:format-wish "menu ~A -tearoff 0~@[ -underline ~a ~]"
		     (ltk:widget-path optionsmenu) underline)
    (loop for x in options for i from 0 do
	  (ltk:format-wish
	   "~a add radiobutton -label ~a -value ~a -variable ~a"
	   (ltk:widget-path optionsmenu) x x  group)
	  if (= i selected-index) do (setf (ltk:value optionsmenu) x))
    optionsmenu))


(defmethod ltk:pack ((optionsmenu optionsmenu) &rest keys &key &allow-other-keys)
  (let ((tk-menubutton (ltk:master optionsmenu)))
    (apply #'ltk:pack tk-menubutton keys)))


;;;
;;;
(defun optionsmenu-index (optionsmenu option)
  "Internal. Retrieve the numeric index of the given OPTION in OPTIONSMENU"
  (ltk:format-wish "senddata [~a index ~a]"  (ltk:widget-path optionsmenu)
		   option)
  (let ((ret (ltk::read-data)))
    (check-type ret (or (member ltk::none) integer))
    ret))

(defun optionsmenu-relabel (optionsmenu index new-option)
  "Internal. Relabel the menu entry at INDEX to NEW-OPTION in OPTIONSMENU"
  (ltk:format-wish "~a entryconfigure ~a -label ~a -value ~a"
		   (ltk:widget-path optionsmenu) index new-option
		   new-option))


;;; ----------------------------------------------------------------------
;;;
;;;  EXAMPLE-1
;;;

#+nil
(ltk:with-ltk ()
  (let* ((sglabels '("Day" "Month" "Week" "Year"))
	 (pllabels '("Days" "Months" "Weeks" "Years"))
	 (numbers :singular)
	 (fr (make-instance 'ltk:frame))
	 (optionsmenu (tk-option-menu :master fr :options sglabels))
	 (vb (make-instance 'ltk:button :master fr :text "Print"
			    :command
			    (lambda ()
			      (warn "VAL=~S" (ltk:value optionsmenu)))))
	 (cb (make-instance
	      'ltk:check-button :master fr :text "Plurals"
	      :onvalue 'on :offvalue 'off
	      :command (lambda (evt)
			 (let ((old-title (ltk:value optionsmenu))
			       (new (ecase evt
				      (on :plural)
				      (off :singular))))
			   (unless (eq numbers new)
			     (loop for i from 0
				   for x in (ecase numbers
					      (:singular sglabels)
					      (:plural pllabels))
				   for y in (ecase new
					      (:singular sglabels)
					      (:plural pllabels))
				   when old-title if (string= old-title x)
				   do (setf (ltk:value optionsmenu) y
					    old-title nil)
				   end
				   do (optionsmenu-relabel optionsmenu i y)
				   finally (setq numbers new)))))))
	 (qb (make-instance 'ltk:button :master fr :text "Quit"
			    :command
			    (lambda () (setq ltk:*EXIT-MAINLOOP* t)))))
    (ltk:configure (ltk:master optionsmenu) :borderwidth 2 :padx 2 :pady 2
		   :highlightthickness 10 :relief :sunken)
    (ltk:pack (list optionsmenu cb vb qb fr))))


;;; ----------------------------------------------------------------------
;;;
;;; EXAMPLE-2   Use of LTK:MENURADIOBUTTONS
;;;

;; LTK ??? perhaps MAKE-INSTANCE should allow a :VALUE initarg and use
;; that in preferance to NAME ???

;;; THE OPTIONSMENU is a suitable menu to add LTK:MENURADIOBUTTONS to.
;;; As of ltk-0.90 the `-value' and `-text' options of the radiobutton
;;; menuentry are taken from the LTK:NAME and LTK:TEXT slots of
;;; LTK:MENURADIOBUTTON respectivelt. Note that changing the slots
;;; will not be reflected in the menuentries without additional
;;; machinery. See below.

#+nil
(ltk:with-ltk ()
  (let* ((fr (make-instance 'ltk:frame))
	 (optionsmenu (tk-option-menu :master fr))
	 (group (optionsmenu-variable optionsmenu))
	 (rb1 (make-instance 'ltk:menuradiobutton :master optionsmenu
			     :group group :name "Day" :text "Day"))
	 (rb2 (make-instance 'ltk:menuradiobutton :master optionsmenu
			     :group group :name "Week" :text "Week"))
	 (rb3 (make-instance 'ltk:menuradiobutton :master optionsmenu
			     :group group :name "Month" :text "Month"
			     :command (lambda ()
					(warn "Month! OK!"))))
	 (vb (make-instance 'ltk:button :master fr :text "Print"
			    :command
			    (lambda ()
			      (warn "VAL=~S" (ltk:value optionsmenu)))))
	 (qb (make-instance 'ltk:button :master fr :text "Quit"
			    :command
			    (lambda () (setq ltk:*EXIT-MAINLOOP* t)))))
    (ltk:configure (ltk:master optionsmenu) :borderwidth 2 :padx 2 :pady 2
		   :highlightthickness 10 :relief :sunken)
    (setf (ltk:value optionsmenu) "[Unselected]")
    (ltk:pack (list optionsmenu vb qb fr))))



;;; ----------------------------------------------------------------------
;;;
;;; OPTIONAL EXTENSIONS TO THE LTK PACKAGE FOR MODIFYING MENUENTRIES
;;; BY SETTING SLOTS ON LTK:MENUBUTTONENTRY (FOR EXAMPLE-3)
;;;

(defmethod entryindex ((menuradiobutton ltk:menuradiobutton))
  "Internal. Retrieve the numeric index of an LTK:MENURADIOBUTTON
amongst the entries of its parent menu."
  (let ((optionsmenu (ltk:master menuradiobutton)))
    (optionsmenu-index optionsmenu (ltk:text menuradiobutton))))


(defmethod (setf menu-value) (val (mrb ltk:menuradiobutton))
  "Sets the `-value' option of the menuentry corresponding to the
given menuradiobutton. NOTE as of LTK-0.90 this option is
initially set from the LTK:NAME slot. This is the same value that
the parent optionsmenu takes if this menuentry is selected."
  (let ((index (entryindex mrb)))
    (assert (numberp index))
    (ltk:format-wish  "~a entryconfigure ~a -value \"~a\""
        (ltk:widget-path (ltk:master mrb)) index
        (ltk::tkescape val))))


(defmethod (setf ltk:text) :before (text (mrb ltk:menuradiobutton))
  "Sets the `-text' option of the menuentry corresponding to the
given menuradiobutton. This is the text displayed in the dropdown
menu.

XXX Currently this method also sets the MENU-VALUE of the
menuradiobutton."
  (let ((index (entryindex mrb)))
    (assert (numberp index))
    ;; XXX assume the (SETF TEXT) generic function call won't fail!
    (setf (menu-value mrb) text) ;; XXX Go at it with a hammer
    (ltk:format-wish "~a entryconfigure ~a -label {~a}"
		     (ltk:widget-path (ltk:master mrb)) index
		     (ltk::tkescape text))))


;;; ----------------------------------------------------------------------
;;;
;;; EXAMPLE-3
;;;
#+nil
(ltk:with-ltk ()
  (let* ((fr (make-instance 'ltk:frame))
	 (optionsmenu (tk-option-menu :master fr))
	 (group (optionsmenu-variable optionsmenu))
	 (rb1 (make-instance 'ltk:menuradiobutton :master optionsmenu
			     :group group :name "Day" :text "Day1"))
	 (rb2 (make-instance 'ltk:menuradiobutton :master optionsmenu
			     :group group :name "Week" :text "Week1"))
	 (rb3 (make-instance 'ltk:menuradiobutton :master optionsmenu
			     :group group :name "Month" :text "Month1"
			     :command (lambda ()
					(warn "Month! OK!"))))
	 (numbers :singular)
	 (cb (make-instance
	      'ltk:check-button :master fr :text "Plurals"
	      :onvalue 'on :offvalue 'off
	      :command
	      (lambda (evt)
		(let ((old-title (ltk:value optionsmenu))
		      (new (ecase evt
			     (on :plural)
			     (off :singular))))
		  (macrolet ((upd (place new-value)
			       `(let ((.oldval. ,place)
				      (.newval. ,new-value))
				  (when old-title
				    (when (string= old-title .oldval.)
				      (setf (ltk:value optionsmenu) .newval.
					    old-title nil)))
				  (setf ,place .newval.))))
		    (upd (ltk:text rb1) (ecase new
					   (:plural "Days1")
					   (:singular "Day2")))
		    (upd (ltk:text rb2) (ecase new
					  (:plural "Weeks1")
					  (:singular "Week2")))
		    (upd (ltk:text rb3) (ecase new
					  (:plural "Months1")
					  (:singular "Month2")))
		    (unless (null old-title)
		      (warn "Sanity check fails. old-title=~S" old-title)))))))
	 (vb (make-instance 'ltk:button :master fr :text "Print"
			    :command
			    (lambda ()
			      (warn "VAL=~S" (ltk:value optionsmenu)))))
	 (qb (make-instance 'ltk:button :master fr :text "Quit"
			    :command
			    (lambda () (setq ltk:*EXIT-MAINLOOP* t)))))
    (ltk:configure (ltk:master optionsmenu) :borderwidth 2 :padx 2 :pady 2
		   :highlightthickness 10 :relief :sunken)
    (setf (ltk:value optionsmenu) "Day1")
    (ltk:pack (list optionsmenu vb cb qb fr))))

'#:FIN
